# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2017 Carlos Jenkins
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Plantuml encoding module.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

import logging
from zlib import compress
from posixpath import join

from requests import get
from six.moves import range
from six import indexbytes, unichr

log = logging.getLogger(__name__)


def compress_and_encode(content):
    """
    Compress the plantuml text and encode it for the plantuml server.

    :param str content: Content to compress and encode.
    :return: The compressed and encoded content.
    :rtype: str
    """
    zlibbed_str = compress(content.strip().encode('utf-8'))
    compressed_string = zlibbed_str[2:-4]
    return encode(compressed_string)


def encode(data):
    """
    Encode given data into PlantUML server encoding.

    This algorithm is similar to the base64 but custom for the plantuml server.

    :param bytes data: Data to encode.
    :return: The encoded data as printable ASCII.
    :rtype: str
    """
    res = ''
    for i in range(0, len(data), 3):
        if i + 2 == len(data):
            res += _encode3bytes(
                indexbytes(data, i),
                indexbytes(data, i + 1),
                0
            )
        elif i + 1 == len(data):
            res += _encode3bytes(
                indexbytes(data, i),
                0, 0
            )
        else:
            res += _encode3bytes(
                indexbytes(data, i),
                indexbytes(data, i + 1),
                indexbytes(data, i + 2)
            )
    return res


def _encode3bytes(b1, b2, b3):
    c1 = b1 >> 2
    c2 = ((b1 & 0x3) << 4) | (b2 >> 4)
    c3 = ((b2 & 0xF) << 2) | (b3 >> 6)
    c4 = b3 & 0x3F
    res = ''
    res += _encode6bit(c1 & 0x3F)
    res += _encode6bit(c2 & 0x3F)
    res += _encode6bit(c3 & 0x3F)
    res += _encode6bit(c4 & 0x3F)
    return res


def _encode6bit(b):
    if b < 10:
        return unichr(48 + b)
    b -= 10
    if b < 26:
        return unichr(65 + b)
    b -= 26
    if b < 26:
        return unichr(97 + b)
    b -= 26
    if b == 0:
        return '-'
    if b == 1:
        return '_'


def plantuml(server, extension, content):
    """
    Call the PlantUML server.

    :param str server: Base URL for the server.
    :param str extension: File format / extension to use for the request.
    :param str content: Content to render.
    :return: Response of the request.
    :rtype: str
    """
    encoded = compress_and_encode(content)
    url = join(server, extension, encoded)
    log.debug('Calling URL:\n{}'.format(url))
    response = get(url)
    response.raise_for_status()
    return response.content


__all__ = ['plantuml']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2017 Carlos Jenkins
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Plantweb main rendering module.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

import logging
from hashlib import sha256
from os import makedirs
from os.path import basename, splitext, isfile, expanduser, join

from .plantuml import plantuml
from .defaults import read_defaults


log = logging.getLogger(__name__)


WRAP_STR = {
    'plantuml': 'uml',
    'ditaa': 'ditaa',
    'graphviz': 'dot'
}
"""
Map between the name of the engine and the wrap string used in the
``@startXXX`` directive.
"""


def determine_engine(content):
    """
    Determine the engine used in the given content.

    The engine is determined by the ``@startXXX`` used at the beginning of the
    file. Possible values are:

    ::

        @startuml --> plantuml
        @startdot --> graphviz
        @startditaa --> ditaa

    :param str content: Content to analyze.
    :return: The name of the engine found, or None.
    :rtype: str
    """
    if content.startswith('@startuml'):
        return 'plantuml'
    if content.startswith('@startdot'):
        return 'graphviz'
    if content.startswith('@startditaa'):
        return 'ditaa'
    return None


def render_cached(
        server, format, content,
        use_cache=None, cache_dir=None):
    """
    Render given content in the PlantUML server or fetch it from cache.

    :param str server: URL to PlantUML server.
    :param str format: File format to render the content. One of the supported
     by the PlantUML server (``svg`` or ``png``).
    :param str content: Content to render with mandatory ``@startxxx`` tags.
    :param bool use_cache: Use local cache to avoid requesting the server for
     already rendered diagrams. If ``None``, the default value will be used.
    :param str cache_dir: Directory to store the cached diagrams. If ``None``
     the default value will be used.

    :return: A tuple of ``(content, sha)`` with the bytes of the rendered
     content and a sha256 hash string identifying the content.
    :rtype: tuple
    """
    sha = sha256(content.encode('utf-8')).hexdigest()

    if use_cache is None:
        use_cache = read_defaults()['use_cache']

    if not use_cache:
        return (
            plantuml(server, format, content),
            sha
        )

    if cache_dir is None:
        cache_dir = read_defaults()['cache_dir']
    cache_dir = expanduser(cache_dir)

    cache_file = join(
        cache_dir, '{}.{}'.format(sha, format)
    )

    # Use cache if available
    log.debug('Trying to load cache file {} ...'.format(cache_file))
    if isfile(cache_file):
        with open(cache_file, 'rb') as fd:
            return (fd.read(), sha)

    # No cache, make sure the directory is available
    makedirs(cache_dir, exist_ok=True)

    # Normal render and save cache
    output = plantuml(server, format, content)

    with open(cache_file, 'wb') as fd:
        fd.write(output)
    log.debug('Wrote cache file {} ...'.format(cache_file))

    return (output, sha)


def render(
        content,
        engine=None,
        format=None,
        server=None,
        cacheopts=None):
    """
    Render given PlantUML, Graphviz or DITAA content.

    :param str content: Content to render.
    :param str engine: Engine to use to render the content. One of
     ``'plantuml'``, ``'graphviz'`` or ``'ditaa'``. If ``None``, the engine
     will be auto-determined by looking into the content for the ``@startxxxx``
     tags, and if unable to be auto-determined the default engine will be used.
    :param str format: Format of the rendered content. Raster ``png`` or vector
     ``svg``. Please note that engine ``ditaa`` can only render to ``png``. If
     ``None``, the default formatwill always be selected unless the engine
     doesn't supports it.
    :param str server: URL to PlantUML server. This will passed as is to
     :func:`render_cached`. If ``None`` the default server URL will be used.

    :return: A tuple of ``(output, format, engine, sha)`` with the bytes of the
     rendered output, a string with the name of the output format, a string
     with the name of the engine used or detected and a string of the sha256
     for identifying the cache file.
    :rtype: tuple
    """

    # Determine engine
    engine_found = determine_engine(content)

    # Case 1: forced engine
    if engine is not None:

        if engine_found is None:
            wrap = WRAP_STR[engine]
            content = '@start{0}\n{1}\n@end{0}'.format(wrap, content)

        elif engine_found != engine:
            log.warning(
                'Engine mismatch. Set: {0} != Found: {1}. '
                'Assuming {1} ...'.format(engine, engine_found)
            )
            engine = engine_found

    # Case 2: Use engine found
    elif engine_found is not None:
        engine = engine_found

    # Case 3: Use a default engine
    else:
        engine = read_defaults()['engine']
        log.warning(
            'Unable to determine the engine. Assuming \'{}\'...'.format(engine)
        )

    # Determine output file format
    if format is None:
        format = read_defaults()['format'] if engine != 'ditaa' else 'png'

    # Determine server
    if server is None:
        server = read_defaults()['server']

    # Render cached
    if cacheopts is None:
        cacheopts = {}

    output, sha = render_cached(server, format, content, **cacheopts)

    return (output, format, engine, sha)


def render_file(infile, outfile=None, renderopts=None, cacheopts=None):
    """
    Render given PlantUML, Graphviz or DITAA file.

    :param str infile: Path to source file to render.
    :param str outfile: Path to output file. If ``None``, the filename will be
     auto-determined and saved to the current working directory.
    :param dict renderopts: Rendering options (``engine``, ``format`` and
     ``server``) as in :func:`render`.
    :param dict cacheopts: Caching options (``use_cache`` and ``cache_dir``) as
     in :func:`render`.

    :return: Path to output file.
    :rtype: str
    """

    # Read source file
    with open(infile, 'rb') as fd:
        content = fd.read().decode('utf-8')

    # Render output
    if renderopts is None:
        renderopts = {}

    output, format, engine, sha = render(
        content, cacheopts=cacheopts, **renderopts
    )

    # Determine destination file
    if outfile is None:
        outfile = '{}.{}'.format(
            splitext(basename(infile))[0],
            format
        )

    # Write output
    with open(outfile, 'wb') as fd:
        fd.write(output)

    return outfile


__all__ = [
    'render_file',
    'render',
    'render_cached',
    'determine_engine',
    'WRAP_STR'
]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2017 Carlos Jenkins
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Sphinx directives for rendering PlantUML, Graphviz and Ditaa using Plantweb.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

from logging import getLogger
from traceback import format_exc
from os import makedirs
from abc import ABCMeta, abstractmethod
from os.path import join, relpath, dirname, isfile, isabs, realpath

from six import add_metaclass
from docutils import nodes
from docutils.parsers.rst import directives
from docutils.parsers.rst.directives.images import Image

from . import defaults
from .render import render


log = getLogger(__name__)


@add_metaclass(ABCMeta)
class Plantweb(Image):
    """
    FIXME: Document.
    """

    required_arguments = 0
    optional_arguments = 1
    option_spec = {
        'alt': directives.unchanged,
        'align': lambda arg: directives.choice(
            arg, ('left', 'center', 'right')
        ),
        'target': directives.unchanged_required,
    }
    has_content = True

    def _directive_checks(self):

        # Check if file insertion is enabled
        if not self.state.document.settings.file_insertion_enabled:
            msg = (
                'File and URL access deactivated. '
                'Ignoring directive "{}".'.format(self._get_directive_name())
            )
            warning = nodes.warning(
                '', self.state_machine.reporter.warning(
                    '', nodes.literal_block('', msg),
                    line=self.lineno
                )
            )
            return [warning]

        # Check that no content and argument are used at the same time
        if self.arguments and self.content:
            warning = self.state_machine.reporter.warning(
                '{} directive cannot have both content and a filename '
                'argument.'.format(self._get_directive_name()),
                line=self.lineno
            )
            return [warning]

        # Check that at least one was provided
        if not (self.arguments or self.content):
            warning = self.state_machine.reporter.warning(
                '{} directive must have content or a filename '
                'argument.'.format(self._get_directive_name()),
                line=self.lineno
            )
            return [warning]

        return None

    def run(self):

        # Execute sanity checks
        warnings = self._directive_checks()
        if warnings:
            return warnings

        # Fetch builder and environment objects
        env = self.state_machine.document.settings.env
        builder = env.app.builder

        # Determine document directory
        document_dir = realpath(dirname(env.doc2path(env.docname)))

        # Load content to render
        if not self.arguments:
            content = '\n'.join(self.content)
        else:
            # Source file should be relative to document, or absolute to
            # configuration directory.
            srcfile = self.arguments[0]

            if isabs(srcfile):
                srcpath = join(env.app.confdir, relpath(srcfile, start='/'))
            else:
                srcpath = join(document_dir, srcfile)

            if not isfile(srcpath):
                warning = self.state_machine.reporter.warning(
                    '{} directive cannot find file {}'.format(
                        self._get_directive_name(),
                        srcfile
                    ),
                    line=self.lineno
                )
                return [warning]

            with open(srcpath, 'rb') as fd:
                content = fd.read().decode('utf-8')

        # Execute plantweb call
        try:
            output, frmt, engine, sha = render(
                content,
                engine=self._get_engine_name()
            )
        except Exception:
            msg = format_exc()
            error = nodes.error(
                '', self.state_machine.reporter.error(
                    '', nodes.literal_block('', msg),
                    line=self.lineno
                )
            )
            return [error]

        # Determine filename
        filename = '{}.{}'.format(sha, frmt)
        imgpath = join(builder.outdir, builder.imagedir, 'plantweb')

        # Create images output folder
        log.debug('imgpath set to {}'.format(imgpath))
        makedirs(imgpath, exist_ok=True)

        # Write content
        filepath = join(imgpath, filename)

        with open(filepath, 'wb') as fd:
            fd.write(output)

        log.debug('Wrote image file {}'.format(filepath))

        # Default to align center
        if 'align' not in self.options:
            self.options['align'] = 'center'

        # Determine relative path to image from source document directory
        filepath_relative = relpath(filepath, document_dir)

        # Windows compatibility:
        # Sphinx Image directive expects paths in POSIX, Python's form.
        # Replace backslash with slash, otherwise they are removed.
        filepath_relative = filepath_relative.replace('\\', '/')

        log.debug('Image relative path {}'.format(filepath_relative))

        # Run Image directive
        self.arguments = [filepath_relative]
        return Image.run(self)

    @abstractmethod
    def _get_engine_name(self):
        """
        Returns the name of the engine that should be used to render the
        content.
        """

    @abstractmethod
    def _get_directive_name(self):
        """
        The name of this directive.
        """


class UmlDirective(Plantweb):
    """
    Specialized ``uml`` directive for Plantweb ``Plantweb`` engine.

    See :class:`Plantweb`.
    """
    def _get_engine_name(self):
        return 'plantuml'

    def _get_directive_name(self):
        return 'uml'


class GraphDirective(Plantweb):
    """
    Specialized ``graph`` directive for Plantweb ``graphviz`` engine.

    See :class:`Plantweb`.
    """
    def _get_engine_name(self):
        return 'graphviz'

    def _get_directive_name(self):
        return 'graph'


class DiagramDirective(Plantweb):
    """
    Specialized ``diagram`` directive for Plantweb ``ditaa`` engine.

    See :class:`Plantweb`.
    """
    def _get_engine_name(self):
        return 'ditaa'

    def _get_directive_name(self):
        return 'diagram'


def defaults_provider():
    """
    Defaults provider that allows to register Sphinx user defaults.

    This dummy defaults provider just returns it's attribute ``overrides`` if
    it exists.

    :return: The dictionary of the form :data:`DEFAULT_CONFIG`.
    :rtype: dict
    """
    return getattr(defaults_provider, 'overrides', {})


def builder_inited_handler(app):
    """
    We use this event handler to grab user defaults for Plantweb and use them
    in Plantweb rendering.

    See https://plantweb.readthedocs.io/index.html#overriding-defaults

    This is the handler of the 'builder-inited' event emitted by Sphinx.

        Emitted when the builder object has been created.
        It is available as app.builder.

    See http://www.sphinx-doc.org/en/stable/extdev/appapi.html#event-builder-inited
    """  # noqa
    log.debug('Sphinx overridden Plantweb defaults:')
    log.debug(app.config.plantweb_defaults)

    # Set overrides in provider
    defaults_provider.overrides = app.config.plantweb_defaults

    # Register provider with the highest priority
    provider = 'python://plantweb.directive.defaults_provider'
    if provider not in defaults.DEFAULTS_PROVIDERS:
        defaults.DEFAULTS_PROVIDERS.append(provider)

    # Force defaults reload
    from .defaults import read_defaults
    if hasattr(read_defaults, 'cache'):
        del read_defaults.cache


def setup(app):
    """
    Setup function that makes this module a Sphinx extension.

    See http://www.sphinx-doc.org/en/stable/extdev/appapi.html#sphinx.application.Sphinx.add_config_value
    """  # noqa
    # Wee want to override the directives:
    # - 'graph' from sphinx.ext.graphviz extension.
    # - 'uml' from sphinxcontrib.plantuml
    # But Sphinx warns of the override, causing failure if warnings are set
    # to fail documentation build. So, we go down and use docutils registering
    # directly instead.

    # app.add_directive('uml', UmlDirective)
    # app.add_directive('graph', GraphDirective)
    # app.add_directive('diagram', DiagramDirective)

    from docutils.parsers.rst import directives
    directives.register_directive('uml', UmlDirective)
    directives.register_directive('graph', GraphDirective)
    directives.register_directive('diagram', DiagramDirective)

    # Register the config value to allow to set plantweb defaults in conf.py
    app.add_config_value('plantweb_defaults', {}, 'env')

    # Register Plantweb defaults setter
    # Note: The str() is because:
    #       - In Python 2.7, Sphinx expects a str, not unicode.
    #       - In Python 3.4, Sphinx expects a str, not bytes.
    app.connect(str('builder-inited'), builder_inited_handler)


__all__ = [
    'Plantweb', 'UmlDirective', 'GraphDirective', 'DiagramDirective',
    'setup', 'builder_inited_handler', 'defaults_provider'
]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2019 KuraLabs S.R.L
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
plantweb executable module entry point.
"""

from sys import exit


def run():

    # Parse arguments
    from .args import InvalidArguments, parse_args
    try:
        args = parse_args()
    except InvalidArguments:
        exit(1)

    # Run program
    from .main import main
    exit(main(args))


if __name__ == '__main__':
    run()


__all__ = []



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2017 Carlos Jenkins
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Plantweb main rendering module.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

import logging
from os import getcwd
from json import loads
from copy import deepcopy
from inspect import isfunction
from traceback import format_exc
from sys import getdefaultencoding
from subprocess import Popen, PIPE
from importlib import import_module
from shutil import which
from os.path import isfile, expanduser, join


log = logging.getLogger(__name__)


DEFAULT_CONFIG = {
    'engine': 'plantuml',
    'format': 'svg',
    'server': 'http://plantuml.com/plantuml/',
    'use_cache': True,
    'cache_dir': '~/.cache/plantweb'
}
"""
Default configuration for plantweb.

.. note::

   The default engine will be used only when the engine was unset and it was
   unable to be auto-determined.

To set a different default configuration create a JSON file ``.plantwebrc``
in your git repository root or in your home, as defined in
:data:`DEFAULTS_PROVIDERS`.
"""


DEFAULTS_PROVIDERS = [
    'python://plantweb.defaults.DEFAULT_CONFIG',
    'file://~/.plantwebrc',
    'git://.plantwebrc'
]
"""
List of defaults providers ordered by read priority.

Last items will be processed last and thus will override previous values.

Available providers are:

``git://``
   Will fetch the repository root from current working directory using git's
   ``git rev-parse --show-toplevel`` and then read the specified file from
   that path.

``file://``
   Will read the file specified. User expansion ``~`` is supported.

``python://``
   Will import the given variable or function:

   - If a function, it will be executed without arguments and its result will
     be used.
   - If a variable, it must be a dictionary similar to :data:`DEFAULT_CONFIG`.
"""


def _read_defaults_git(path):
    """
    Read defaults from given path in current git repository.

    See :data:`DEFAULTS_PROVIDERS` for inner workings.
    """
    # Find git executable
    git = which('git')
    if not git:
        log.debug(
            'Unable to read defaults from repository. '
            'git executable not found.'
        )
        return {}

    # Call git to determine repository root
    proc = Popen(
        [git, 'rev-parse', '--show-toplevel'],
        stdout=PIPE, stderr=PIPE
    )
    stdout_raw, stderr_raw = proc.communicate()
    stdout = stdout_raw.decode(getdefaultencoding())
    stderr = stderr_raw.decode(getdefaultencoding())

    # Check that we were in a repository
    if proc.returncode != 0:
        if 'not a git repository' in stderr.lower():
            log.debug(
                'Not in a git repository: {}'.format(getcwd())
            )
        else:
            log.error(
                'Unable to determine git root directory:\n{}'.format(stderr)
            )
        return {}

    # Read defaults file
    repo_root = stdout.strip()

    return _read_defaults_file(
        join(repo_root, path)
    )


def _read_defaults_file(path):
    """
    Read defaults from given path.

    See :data:`DEFAULTS_PROVIDERS` for inner workings.
    """

    rcfile = expanduser(path)
    if isfile(rcfile):
        with open(rcfile, 'r') as fd:
            return loads(fd.read())

    log.info('Defaults file {} doesn\'t exists'.format(rcfile))
    return {}


def _read_defaults_python(location):
    """
    Import and read defaults from given Python entity.

    See :data:`DEFAULTS_PROVIDERS` for inner workings.
    """
    try:
        module_path, variable_name = location.rsplit('.', 1)
        module = import_module(module_path)

        if not hasattr(module, variable_name):
            log.warning('Unknown entity {}'.format(location))
            return {}

        entity = getattr(module, variable_name)

        if isfunction(entity):
            return entity()

        return entity

    except Exception:
        log.debug(format_exc())
        log.warning('Unable to load entity {}'.format(location))

    return {}


SCHEMES = {
    'git://': _read_defaults_git,
    'file://': _read_defaults_file,
    'python://': _read_defaults_python
}


def read_defaults(cached=True):
    """
    Get the defaults values.

    :param bool cached: Read cached default values or determine them from
     the list of providers. See :data:`DEFAULTS_PROVIDERS`.

    :return: A dictionary like :data:`DEFAULT_CONFIG` with the user defaults.
    :rtype: dict
    """
    if cached and hasattr(read_defaults, 'cache'):
        return read_defaults.cache

    defaults = {}

    for provider in DEFAULTS_PROVIDERS:
        for scheme, func in SCHEMES.items():
            if provider.startswith(scheme):
                overrides = func(
                    provider.replace(scheme, '', 1)
                )
                defaults.update(deepcopy(overrides))

                break
        else:
            raise Exception('Unknown provider URI {}'.format(provider))

    read_defaults.cache = defaults
    return defaults


__all__ = ['DEFAULT_CONFIG', 'DEFAULTS_PROVIDERS', 'read_defaults']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2017 Carlos Jenkins
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Argument management module.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

import logging
from os import makedirs
from os.path import isfile, abspath, expanduser

from . import __version__


log = logging.getLogger(__name__)


FORMAT = '%(asctime)s:::%(levelname)s:::%(message)s'
V_LEVELS = {
    0: logging.ERROR,
    1: logging.WARNING,
    2: logging.INFO,
    3: logging.DEBUG,
}


class InvalidArguments(Exception):
    """
    Typed exception that allows to fail in argument parsing and verification
    without quiting the process.
    """
    pass


def validate_args(args):
    """
    Validate that arguments are valid.

    :param args: An arguments namespace.
    :type args: :py:class:`argparse.Namespace`
    :return: The validated namespace.
    :rtype: :py:class:`argparse.Namespace`
    """
    level = V_LEVELS.get(args.verbose, logging.DEBUG)
    logging.basicConfig(format=FORMAT, level=level)

    log.debug('Raw arguments:\n{}'.format(args))

    # Verify source file exists
    sources = []
    for src in args.sources:
        if not isfile(src):
            raise InvalidArguments(
                'No such file {}'.format(src)
            )
        sources.append(abspath(src))
    args.sources = sources

    # Check that format and engine compatibility
    if args.format == 'svg' and args.engine == 'ditaa':
        raise InvalidArguments(
            'The ditaa engine doens\'t support the svg format'
        )

    # Prepare default datatypes
    if args.engine == 'auto':
        args.engine = None
    if args.format == 'auto':
        args.format = None

    # Ensure cache dir
    if not args.no_cache:
        makedirs(expanduser(args.cache_dir), exist_ok=True)

    return args


def parse_args(argv=None):
    """
    Argument parsing routine.

    :param argv: A list of argument strings.
    :rtype argv: list
    :return: A parsed and verified arguments namespace.
    :rtype: :py:class:`argparse.Namespace`
    """
    from argparse import ArgumentParser

    parser = ArgumentParser(
        description='Python client for the PlantUML server'
    )
    parser.add_argument(
        '-v', '--verbose',
        help='increase verbosity level',
        default=0,
        action='count'
    )
    parser.add_argument(
        '--version',
        action='version',
        version='Plantweb {}'.format(__version__)
    )

    parser.add_argument(
        '--engine',
        default='auto',
        help='engine to use to render diagram',
        choices=['auto', 'plantuml', 'graphviz', 'ditaa']
    )
    parser.add_argument(
        '--format',
        default='auto',
        help='diagram export format',
        choices=['auto', 'svg', 'png']
    )

    parser.add_argument(
        '--server',
        default='http://plantuml.com/plantuml/',
        help='server to use for rendering'
    )

    parser.add_argument(
        '--no-cache',
        action='store_true',
        help='do not use cache'
    )
    parser.add_argument(
        '--cache-dir',
        default='~/.cache/plantweb',
        help='directory to store cached renders'
    )

    parser.add_argument(
        'sources',
        nargs='+',
        help='source files to render'
    )

    args = parser.parse_args(argv)
    try:
        args = validate_args(args)
    except InvalidArguments as e:
        log.critical(e)
        raise e

    return args


__all__ = ['parse_args']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2017 Carlos Jenkins
# Copyright (C) 2019 KuraLabs S.R.L
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from setuptools import setup, find_packages


def read(filename):
    """
    Read a file relative to setup.py location.
    """
    import os
    here = os.path.dirname(os.path.abspath(__file__))
    with open(os.path.join(here, filename)) as fd:
        return fd.read()


def find_version(filename):
    """
    Find package version in file.
    """
    import re
    content = read(filename)
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]", content, re.M
    )
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')


def find_requirements(filename):
    """
    Find requirements in file.
    """
    import string
    content = read(filename)
    requirements = []
    for line in content.splitlines():
        line = line.strip()
        if line and line[:1] in string.ascii_letters:
            requirements.append(line)
    return requirements


setup(
    name='plantweb',
    version=find_version('lib/plantweb/__init__.py'),
    package_dir={'': 'lib'},
    packages=find_packages('lib'),

    # Dependencies
    install_requires=find_requirements('requirements.txt'),

    # Metadata
    author='Carlos Jenkins',
    author_email='carlos@jenkins.co.cr',
    description=(
        'Python client for the PlantUML server'
    ),
    long_description=read('README.rst'),
    url='https://plantweb.rtfd.io/',
    keywords='plantweb',

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
    ],
    # Sphinx autodoc cannot extract the documentation of zipped eggs with the
    # ``.. autodata::`` directive, causing caos in autoapi.
    # With this, we will disabled zip_safe flag for Readthedocs for correct
    # autoapi generation.
    zip_safe=False,

    entry_points={
        'console_scripts': ['plantweb=plantweb.__main__:run'],
    },
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2017 Carlos Jenkins
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Application entry point module.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

import logging

from .render import render_file


log = logging.getLogger(__name__)


def main(args):
    """
    Application main function.

    :param args: An arguments namespace.
    :type args: :py:class:`argparse.Namespace`

    :return: Exit code.
    :rtype: int
    """
    for src in args.sources:

        destination = render_file(
            src,
            renderopts={
                'engine': args.engine,
                'format': args.format,
                'server': args.server
            },
            cacheopts={
                'use_cache': not args.no_cache,
                'cache_dir': args.cache_dir
            }
        )

        print('Writing output for {} to {}'.format(src, destination))

    return 0


__all__ = [
    'main',
]
